from typing import Any
from typing import List
from typing import Optional

REPORT_FILTER_DB_NAME = "report_filter"
REPORT_FILTER_DB_DIRNAME = "report_filters"
REPORT_FILTER_REGISTRY_NAME = "registry"
DEFAULT_FILTER_NAME = "default"
SAMPLE_TO_FILTER_TABLE_NAME = "fsamplegiven"
GENE_TO_FILTER_TABLE_NAME = "fgenegiven"
REPORT_FILTER_IN_PROGRESS = "in_progress"
REPORT_FILTER_READY = "ready"
REPORT_FILTER_NOT_NEEDED = "not_needed"
REF_COL_NAMES = {
    "variant": "base__uid",
    "gene": "base__hugo",
    "sample": "base__uid",
    "mapping": "base__uid",
}
level_prefixes = {"variant": "v", "gene": "g"}


class FilterColumn(object):

    test2sql = {
        "equals": "==",
        "lessThanEq": "<=",
        "lessThan": "<",
        "greaterThanEq": ">=",
        "greaterThan": ">",
        "hasData": "is not null",
        "noData": "is null",
        "stringContains": "like",
        "stringStarts": "like",  # Deprecated. Eliminate later
        "stringEnds": "like",  # Deprecated. Eliminate later
        "between": "between",
        "in": "in",
        "select": "in",
    }

    def __init__(self, d, parent_operator):
        self.column = d["column"]
        if self.column == "base__numsample":
            self.column = "tagsampler__numsample"
        self.test = d["test"]
        self.value = d.get("value")
        self.negate = d.get("negate", False)
        self.parent_operator = parent_operator
        self.level = d.get("level")

    def __repr__(self):
        return f"{self.column} {self.test} {self.value}"

    def get_sql(self):
        s = ""
        col_name = f"{level_prefixes[self.level]}.{self.column}"
        if self.test == "multicategory":
            s = '{} like "%{}%"'.format(col_name, self.value[0])
            for v in self.value[1:]:
                s += ' or {} like "%{}%"'.format(col_name, v)
        elif self.test in ("select", "in"):
            ss = []
            for val in self.value:
                if type(val) == str:
                    val = '"{}"'.format(val)
                else:
                    val = str(val)
                ss.append(f"({col_name} = {val})")
            if ss:
                s = "(" + f" or ".join(ss) + ")"
            else:
                s = ""
        else:
            s = "{col} {opr}".format(col=col_name, opr=self.test2sql[self.test])
            sql_val = None
            if self.test == "equals":
                if type(self.value) is list:
                    v = self.value[0]
                    if type(v) is str:
                        sql_val = '"' + v + '"'
                    else:
                        sql_val = str(v)
                    for v in self.value[1:]:
                        if type(v) is str:
                            v = '"' + v + '"'
                        else:
                            v = str(v)
                        sql_val += " OR {} == {}".format(col_name, v)
                else:
                    if type(self.value) is str:
                        sql_val = '"{}"'.format(self.value)
                    else:
                        sql_val = str(self.value)
            elif self.test == "stringContains":
                sql_val = '"%{}%"'.format(self.value)
            elif self.test == "stringStarts":
                sql_val = '"{}%"'.format(self.value)
            elif self.test == "stringEnds":
                sql_val = '"%{}"'.format(self.value)
            elif self.test == "between":
                sql_val = "{} and {}".format(self.value[0], self.value[1])
            elif self.test in (
                "lessThan",
                "lessThanEq",
                "greaterThan",
                "greaterThanEq",
            ):
                sql_val = str(self.value)
            if sql_val:
                s += " (" + sql_val + ")"
        if self.negate:
            s = "not(" + s + ")"
        return s


class FilterGroup(object):
    def __init__(self, d):
        self.operator = d.get("operator", "and")
        self.negate = d.get("negate", False)
        self.rules = []
        for rule in d.get("rules", []):
            if "operator" in rule:
                self.rules.append(FilterGroup(rule))
            else:
                self.rules.append(FilterColumn(rule, self.operator))
        # Backwards compatability, may remove later
        self.rules += [FilterGroup(x) for x in d.get("groups", [])]
        self.rules += [FilterColumn(x, self.operator) for x in d.get("columns", [])]

    def get_sql(self):
        clauses = []
        for operand in self.rules:
            clause = operand.get_sql()
            if clause:
                clauses.append(clause)
        s = ""
        if clauses:
            s += "("
            sql_operator = " " + self.operator + " "
            s += sql_operator.join(clauses)
            s += ")"
            if self.negate:
                s = "not" + s
        return s


class ReportFilter:
    from ..system.consts import DEFAULT_SERVER_DEFAULT_USERNAME

    @classmethod
    async def create(
        cls,
        dbpath=None,
        filterpath=None,
        filtername=None,
        filterstring=None,
        filter=None,
        mode="sub",
        filtersql=None,
        includesample=None,
        excludesample=None,
        strict=True,
        user=DEFAULT_SERVER_DEFAULT_USERNAME,
        uid=None,
    ):
        self = ReportFilter(
            dbpath=dbpath,
            filterpath=filterpath,
            filtername=filtername,
            filterstring=filterstring,
            filter=filter,
            mode=mode,
            filtersql=filtersql,
            includesample=includesample,
            excludesample=excludesample,
            strict=strict,
            user=user,
            uid=uid,
        )
        await self.second_init()
        return self

    def __init__(
        self,
        dbpath=None,
        filterpath=None,
        filtername=None,
        filterstring=None,
        filter=None,
        filtersql=None,
        includesample=None,
        excludesample=None,
        mode="sub",
        strict=True,
        user=DEFAULT_SERVER_DEFAULT_USERNAME,
        uid=None,
    ):
        from pathlib import Path

        self.mode = mode
        if self.mode == "main":
            self.stdout = True
        else:
            self.stdout = False
        self.dbpath = dbpath
        if self.dbpath:
            self.dbpath = str(Path(dbpath).absolute())
        self.conn = None
        self.filterpath = filterpath
        self.cmd = None
        self.level = None
        self.filter = filter
        self.filtername = None
        self.filterstring = filterstring
        self.filtersql = filtersql
        self.includesample = includesample
        self.excludesample = excludesample
        self.strict = strict
        if filter != None:
            self.filter = filter
        else:
            if filterstring != None:
                self.filterstring = filterstring
            elif filtername != None:
                self.filtername = filtername
            elif filterpath != None:
                self.filterpath = filterpath
        self.filtertable = "filter"
        self.generows = {}
        self.table_aliases = None
        self.conn_read = None
        self.conn_write = None
        self.uid = uid
        self.user = self.escape_user(user)

    async def exec_db(self, func, *args, **kwargs):
        if not self.conn_read or not self.conn_write:
            return None
        cursor_read = await self.conn_read.cursor()
        cursor_write = await self.conn_write.cursor()
        ret = await func(
            *args, cursor_read=cursor_read, cursor_write=cursor_write, **kwargs
        )
        await cursor_read.close()
        await cursor_write.close()
        return ret

    async def second_init(self):
        if self.mode == "sub":
            if self.dbpath != None:
                await self.connect_dbs()
            await self.exec_db(self.loadfilter)

    # async def run_level_based_func(self, cmd):
    #    ret = {}
    #    if self.level != None:
    #        ret[self.level] = await cmd(level=self.level)
    #    else:
    #        levels = ["variant", "gene"]
    #        ret = {}
    #        for level in levels:
    #            ret_onelevel = await cmd(level=level)
    #            ret[level] = ret_onelevel
    #    return ret

    def parse_args(self, args):
        from argparse import ArgumentParser
        from os.path import abspath

        parser = ArgumentParser()
        parser.add_argument(
            "-d",
            dest="dbpath",
            required=True,
            help="Path of a result database file (.sqlite)",
        )
        parser.add_argument(
            "-f", dest="filterpath", help="Path of a filtering criteria file"
        )
        parser.add_argument(
            "-F",
            dest="filtername",
            help="Name of the filter to apply (saved in the database)",
        )
        parser.add_argument(
            "--filterstring", dest="filterstring", default=None, help="Filter in JSON"
        )
        parser.add_argument(
            "-l",
            dest="level",
            default=None,
            choices=["variant", "gene"],
            help="Analysis level to filter",
        )
        parser.add_argument(
            "--filtersql", dest="filtersql", default=None, help="Filter SQL"
        )
        parser.add_argument(
            "--includesample",
            dest="includesample",
            nargs="+",
            default=None,
            help="Sample IDs to include",
        )
        parser.add_argument(
            "--excludesample",
            dest="excludesample",
            nargs="+",
            default=None,
            help="Sample IDs to exclude",
        )
        if self.mode == "main":
            parser.add_argument(
                "command",
                choices=["uidpipe", "count", "rows", "pipe", "save", "list"],
                help="Command",
            )

        parsed_args = parser.parse_args(args)
        self.dbpath = abspath(parsed_args.dbpath)
        self.filterpath = parsed_args.filterpath
        self.level = parsed_args.level
        if self.mode == "main":
            self.cmd = parsed_args.command
        self.filtername = parsed_args.filtername
        self.filterstring = parsed_args.filterstring
        self.filtersql = parsed_args.filtersql

    def get_report_filter_db_dir(self):
        from ..system import get_user_conf_dir
        from os.path import join

        user_conf_dir = get_user_conf_dir()
        return join(user_conf_dir, REPORT_FILTER_DB_DIRNAME)

    def escape_user(self, user):
        return "".join([c if c.isalnum() else "_" for c in user])

    def get_report_filter_db_fn(self):
        report_filter_db_fn = f"report_filter.{self.user}.sqlite"
        return report_filter_db_fn

    def get_report_filter_db_path(self):
        from os.path import join

        report_filter_db_dir = self.get_report_filter_db_dir()
        report_filter_db_fn = self.get_report_filter_db_fn()
        return join(report_filter_db_dir, report_filter_db_fn)

    def get_registry_table_name(self):
        return f"{REPORT_FILTER_DB_NAME}.{REPORT_FILTER_REGISTRY_NAME}"

    async def create_report_filter_registry_table_if_not_exists(self, conn):
        cursor = await conn.cursor()
        q = f"create table if not exists {REPORT_FILTER_DB_NAME}.{REPORT_FILTER_REGISTRY_NAME} ( uid int primary key, user text, dbpath text, filterjson text, status text )"
        await cursor.execute(q)
        await conn.commit()
        await cursor.close()

    async def create_and_attach_filter_database(self, conn):
        from os.path import exists
        from os import mkdir

        if not conn:
            return
        cursor = await conn.cursor()
        report_filter_db_dir = self.get_report_filter_db_dir()
        if not exists(report_filter_db_dir):
            mkdir(report_filter_db_dir)
        report_filter_db_path = self.get_report_filter_db_path()
        q = f"attach database '{report_filter_db_path}' as {REPORT_FILTER_DB_NAME}"
        await cursor.execute(q)
        await cursor.close()
        await self.create_report_filter_registry_table_if_not_exists(conn)

    async def make_db_conns(self):
        from aiosqlite import connect
        from aiosqlite import Row

        if not self.dbpath:
            return None
        if not self.conn_read:
            self.conn_read = await connect(self.dbpath)
            self.conn_read.row_factory = Row
            await self.conn_read.execute("pragma journal_mode=wal")
            await self.create_and_attach_filter_database(self.conn_read)
        if not self.conn_write:
            self.conn_write = await connect(self.dbpath)
            await self.conn_write.execute("pragma journal_mode=wal")
            await self.create_and_attach_filter_database(self.conn_write)

    async def connect_dbs(self, dbpath=None):
        from os.path import abspath

        if dbpath != None:
            self.dbpath = abspath(dbpath)
        await self.make_db_conns()

    async def close_db(self):
        if self.conn_read:
            await self.conn_read.close()
            self.conn_read = None
        if self.conn_write:
            await self.conn_write.close()
            self.conn_write = None

    async def filtertable_exists(self, cursor_read=Any, cursor_write=Any):
        _ = cursor_write
        sql = f'select * from viewersetup where datatype="filter" and name="{self.filtername}"'
        await cursor_read.execute(sql)
        row = await cursor_read.fetchone()
        if row == None:
            ret = False
        else:
            ret = True
        return ret

    async def loadfilter(
        self,
        filterpath=None,
        filtername=None,
        filterstring=None,
        filtersql=None,
        filter=None,
        includesample=None,
        excludesample=None,
        cursor_read=Any,
        cursor_write=Any,
        strict=None,
    ):
        from os.path import exists
        from oyaml import safe_load
        from json import load, loads

        _ = cursor_write
        if strict:
            self.strict = strict
        if filterpath != None:
            self.filterpath = filterpath
        if filtername != None:
            self.filtername = filtername
        if filterstring != None:
            self.filterstring = filterstring
        if filtersql != None:
            self.filtersql = filtersql
        if filter != None:
            self.filter = filter
        if includesample is not None:
            self.includesample = includesample
        if excludesample is not None:
            self.excludesample = excludesample
        filter_table_present = await self.exec_db(self.filtertable_exists)
        if self.filter:
            pass
        elif self.filtersql is not None:
            if exists(self.filtersql):  # a file can be used.
                with open(self.filtersql) as f:
                    self.filtersql = "".join(f.readlines())
            self.filter = {}
        elif self.filterstring is not None:
            self.filterstring = self.filterstring.replace("'", '"')
            self.filter = loads(self.filterstring)
        elif self.filtername is not None and filter_table_present:
            await cursor_read.execute(
                "select viewersetup from viewersetup"
                + ' where datatype="filter" and name="'
                + self.filtername
                + '"'
            )
            criteria = await cursor_read.fetchone()
            if criteria != None:
                self.filter = loads(criteria[0])["filterSet"]
        elif self.filterpath is not None and exists(self.filterpath):
            with open(self.filterpath) as f:
                ftype = self.filterpath.split(".")[-1]
                if ftype in ["yml", "yaml"]:
                    self.filter = safe_load(f)
                elif ftype in ["json"]:
                    self.filter = load(f)
        if self.filter is None:
            self.filter = {}
        await self.exec_db(self.verify_filter)

    async def verify_filter(self, cursor_read=Any, cursor_write=Any):
        from oakvar.exceptions import InvalidFilter

        _ = cursor_write
        wrong_samples = await self.verify_filter_sample(cursor_read)
        wrong_colnames = await self.verify_filter_module(cursor_read)
        if self.strict and (len(wrong_samples) > 0 or len(wrong_colnames) > 0):
            raise InvalidFilter(wrong_samples, wrong_colnames)

    async def check_sample_name(self, sample_id, cursor):
        await cursor.execute(
            'select base__sample_id from sample where base__sample_id="'
            + sample_id
            + '" limit 1'
        )
        ret = await cursor.fetchone()
        return ret is not None

    async def verify_filter_sample(self, cursor):
        if not self.filter or "sample" not in self.filter:
            return []
        ft = self.filter["sample"]
        wrong_samples = set()
        if "require" in ft:
            for rq in ft["require"]:
                if await self.check_sample_name(rq, cursor) == False:
                    wrong_samples.add(rq)
        if not self.strict:
            for s in wrong_samples:
                if s in ft["require"]:
                    ft["require"].remove(s)
        if "reject" in ft:
            for rq in ft["reject"]:
                if await self.check_sample_name(rq, cursor) == False:
                    wrong_samples.add(rq)
        if not self.strict:
            for s in wrong_samples:
                if s in ft["reject"]:
                    ft["reject"].remove(s)
        return wrong_samples

    async def col_name_exists(self, colname, cursor):
        await cursor.execute(
            'select col_def from variant_header where col_name="'
            + colname
            + '" limit 1'
        )
        ret = await cursor.fetchone()
        if ret is None:
            await cursor.execute(
                'select col_def from gene_header where col_name="'
                + colname
                + '" limit 1'
            )
            ret = await cursor.fetchone()
        return ret is not None

    async def extract_filter_columns(self, rules, colset, cursor):
        rules_to_delete = []
        for rule in rules:
            if "column" in rule:
                if await self.col_name_exists(rule["column"], cursor) == False:
                    colset.add(rule["column"])
                    rules_to_delete.append(rule)
            elif "rules" in rule:
                await self.extract_filter_columns(rule["rules"], colset, cursor)
        if not self.strict:
            for rule in rules_to_delete:
                rules.remove(rule)

    async def verify_filter_module(self, cursor):
        if "variant" not in self.filter:
            return []
        wrong_modules = set()
        if "rules" in self.filter["variant"]:
            await self.extract_filter_columns(
                self.filter["variant"]["rules"], wrong_modules, cursor
            )
        return wrong_modules

    def getwhere(self, level):
        where = ""
        if self.filter and level in self.filter:
            criteria = self.filter[level]
            main_group = FilterGroup(criteria)
            sql = main_group.get_sql()
            if sql:
                where = "where " + sql
        return where

    def get_sample_to_filter(self):
        if not self.filter:
            return None
        req = []
        rej = []
        if "sample" in self.filter:
            if "require" in self.filter["sample"]:
                req = self.filter["sample"]["require"]
            if "reject" in self.filter["sample"]:
                rej = self.filter["sample"]["reject"]
        if self.includesample:
            req = self.includesample
        if self.excludesample:
            rej = self.excludesample
        if len(req) == 0 and len(rej) == 0:
            return None
        return [req, rej]

    def get_sample_to_filter_table_name(self, uid=None):
        if not uid:
            return None
        return f"{REPORT_FILTER_DB_NAME}.{SAMPLE_TO_FILTER_TABLE_NAME}_{uid}"

    async def remove_temporary_tables(
        self,
        uid=None,
        cursor_read=Any,
        cursor_write=Any,
        gene_to_filter=None,
        sample_to_filter=None,
    ):
        if not self.conn_write:
            return
        _ = cursor_read
        if sample_to_filter:
            table_name = self.get_sample_to_filter_table_name(uid=uid)
            if not table_name:
                return
            q = f"drop table if exists {table_name}"
            await cursor_write.execute(q)
        if gene_to_filter:
            table_name = self.get_gene_to_filter_table_name(uid=uid)
            if not table_name:
                return
            q = f"drop table if exists {table_name}"
            await cursor_write.execute(q)
        await self.conn_write.commit()

    async def make_sample_to_filter_table(
        self, uid=None, req=None, rej=None, cursor_read=Any, cursor_write=Any
    ):
        if not self.conn_write:
            return
        _ = cursor_read
        if not uid or (not req and not rej):
            return
        table_name = self.get_sample_to_filter_table_name(uid=uid)
        if not table_name:
            return
        q = f"drop table if exists {table_name}"
        await cursor_write.execute(q)
        await self.conn_write.commit()
        q = f"create table {table_name} as select distinct base__uid from main.sample"
        if req:
            req_s = ", ".join([f'"{sid}"' for sid in req])
            q += f" where base__sample_id in ({req_s})"
        if rej:
            rej_s = ", ".join([f'"{sid}"' for sid in rej])
            q += f" except select base__uid from main.sample where base__sample_id in ({rej_s})"
        await cursor_write.execute(q)
        await self.conn_write.commit()

    async def get_existing_report_filter_status(
        self, cursor_read=Any, cursor_write=Any
    ):
        from json import dumps

        _ = cursor_write
        if not self.filter or not self.dbpath or not cursor_read:
            return None
        filterjson = dumps(self.filter)
        tablename = self.get_registry_table_name()
        q = f"select uid, status from {tablename} where user=? and dbpath=? and filterjson=?"
        await cursor_read.execute(q, (self.user, self.dbpath, filterjson))
        ret = await cursor_read.fetchone()
        if not ret:
            return None
        [uid, status] = ret
        return {"uid": uid, "status": status}

    async def get_report_filter_count(self, cursor=Any):
        tablename = self.get_registry_table_name()
        q = f"select count(*) from {tablename}"
        await cursor.execute(q)
        ret = await cursor.fetchone()
        count = ret[0]
        return count

    async def get_max_report_filter_uid(self, cursor=Any, where=None):
        tablename = self.get_registry_table_name()
        q = f"select max(uid) from {tablename}"
        if where:
            q += f" where {where}"
        await cursor.execute(q)
        ret = await cursor.fetchone()
        n = ret[0]
        return n

    async def get_min_report_filter_uid(self, cursor=Any, where=None):
        tablename = self.get_registry_table_name()
        q = f"select min(uid) from {tablename}"
        if where:
            q += f" where {where}"
        await cursor.execute(q)
        ret = await cursor.fetchone()
        n = ret[0]
        return n

    async def get_report_filter_string(
        self, uid=None, cursor_read=Any, cursor_write=Any
    ) -> Optional[str]:
        if not uid:
            return None
        _ = cursor_write
        tablename = self.get_registry_table_name()
        q = f"select filterjson from {tablename} where uid=?"
        await cursor_read.execute(q, (uid,))
        ret = await cursor_read.fetchone()
        if ret:
            return ret[0]
        else:
            return ret

    async def get_new_report_filter_uid(self, cursor_read=Any, cursor_write=Any):
        from ..system import get_sys_conf_value
        from ..system.consts import report_filter_max_num_cache_per_user_key
        from ..system.consts import DEFAULT_REPORT_FILTER_MAX_NUM_CACHE_PER_USER

        _ = cursor_write
        delete_uids = []
        report_filter_max_num_cache_per_user = get_sys_conf_value(
            report_filter_max_num_cache_per_user_key
        )
        if not report_filter_max_num_cache_per_user:
            report_filter_max_num_cache_per_user_key = (
                DEFAULT_REPORT_FILTER_MAX_NUM_CACHE_PER_USER
            )
        count = await self.get_report_filter_count(cursor=cursor_read)
        max_uid = await self.get_max_report_filter_uid(cursor=cursor_read)
        min_uid = await self.get_min_report_filter_uid(cursor=cursor_read)
        if count < report_filter_max_num_cache_per_user:
            uid = count
        else:
            if max_uid < report_filter_max_num_cache_per_user * 2 - 1:
                delete_uids = [min_uid]
                uid = max_uid + 1
            else:
                delete_uids = [
                    v
                    for v in range(
                        report_filter_max_num_cache_per_user * 2, max_uid + 1
                    )
                ]
                min_over_uid = await self.get_min_report_filter_uid(
                    cursor=cursor_read,
                    where=f"uid >= {report_filter_max_num_cache_per_user}",
                )
                delete_uids.append(min_over_uid)
                uid = min_over_uid - report_filter_max_num_cache_per_user
        return [uid, delete_uids]

    async def get_filtered_hugo_list(self, cursor_read=Any, cursor_write=Any):
        _ = cursor_write
        if self.should_bypass_filter():
            table_name = "main.gene"
        else:
            table_name = self.get_ftable_name(uid=self.uid, ftype="gene")
        q = f"select base__hugo from {table_name}"
        await cursor_read.execute(q)
        rets = [v[0] for v in await cursor_read.fetchall()]
        return rets

    async def register_new_report_filter(
        self, uid: int, cursor_read=Any, cursor_write=Any
    ):
        from json import dumps

        if not self.conn_write:
            return
        _ = cursor_read
        filterjson = dumps(self.filter)
        q = f"insert or replace into {REPORT_FILTER_DB_NAME}.{REPORT_FILTER_REGISTRY_NAME} ( uid, user, dbpath, filterjson, status) values (?, ?, ?, ?, ?)"
        await cursor_write.execute(
            q, (uid, self.user, self.dbpath, filterjson, REPORT_FILTER_IN_PROGRESS)
        )
        await self.conn_write.commit()

    def should_bypass_filter(self):
        return (
            not self.filter
            and not self.filtersql
            and not self.includesample
            and not self.excludesample
        )

    def get_ftable_name(self, uid=None, ftype=None):
        if not uid or not ftype:
            return None
        return f"{REPORT_FILTER_DB_NAME}.f{ftype}_{uid}"

    def get_fvariant_sql(self, uid=None, gene_to_filter=None, sample_to_filter=None):
        q = f"select v.base__uid from main.variant as v"
        if not uid:
            return q
        if gene_to_filter:
            gene_to_filter_table_name = self.get_gene_to_filter_table_name(uid=uid)
            q += (
                f" join {gene_to_filter_table_name} as gl on v.base__hugo=gl.base__hugo"
            )
        if sample_to_filter:
            sample_to_filter_table_name = self.get_sample_to_filter_table_name(uid=uid)
            q += (
                f" join {sample_to_filter_table_name} as sl on v.base__uid=sl.base__uid"
            )
        if self.filter:
            where = self.getwhere("variant")
            if "g." in where:
                q += " join gene as g on v.base__hugo=g.base__hugo"
            q += " " + where
        elif self.filtersql:
            if "g." in self.filtersql:
                q += f" join main.gene as g on v.base__hugo=g.base__hugo"
            if "s." in self.filtersql:
                q += f" join main.sample as s on v.base__uid=s.base__uid"
            q += " where " + self.filtersql
        # q += " order by base__uid"
        return q

    async def populate_fvariant(
        self,
        uid=None,
        sample_to_filter=None,
        gene_to_filter=None,
        cursor_read=Any,
        cursor_write=Any,
    ):
        _ = cursor_read
        if not uid or not self.conn_write:
            return
        table_name = self.get_ftable_name(uid=uid, ftype="variant")
        q = self.get_fvariant_sql(
            uid=uid, gene_to_filter=gene_to_filter, sample_to_filter=sample_to_filter
        )
        q = f"create table {table_name} as {q}"
        await cursor_write.execute(q)
        await self.conn_write.commit()

    async def populate_fgene(self, uid=None, cursor_read=Any, cursor_write=Any):
        if not uid or not self.conn_write:
            return
        _ = cursor_read
        table_name = self.get_ftable_name(uid=uid, ftype="gene")
        fvariant = self.get_ftable_name(uid=uid, ftype="variant")
        q = f"create table {table_name} as select distinct v.base__hugo from main.variant as v inner join {fvariant} as vf on vf.base__uid=v.base__uid where v.base__hugo is not null"
        await cursor_write.execute(q)
        await self.conn_write.commit()

    async def make_fvariant(self, uid=None, sample_to_filter=None, gene_to_filter=None):
        if not uid:
            return
        await self.exec_db(self.drop_ftable, uid=uid, ftype="variant")
        await self.exec_db(
            self.populate_fvariant,
            uid=uid,
            gene_to_filter=gene_to_filter,
            sample_to_filter=sample_to_filter,
        )

    async def make_fgene(self, uid=None):
        if not uid:
            return
        await self.exec_db(self.drop_ftable, uid=uid, ftype="gene")
        await self.exec_db(self.populate_fgene, uid=uid)

    async def set_registry_status(
        self, uid=None, status=None, cursor_read=Any, cursor_write=Any
    ):
        _ = cursor_read
        if not uid or not status or not self.conn_write:
            return
        table_name = self.get_registry_table_name()
        q = f"update {table_name} set status=?"
        await cursor_write.execute(q, (status,))
        await self.conn_write.commit()

    async def remove_ftables(self, uids, cursor_read=Any, cursor_write=Any):
        _ = cursor_read
        if not self.conn_write:
            return
        if type(uids) == int:
            uids = [uids]
        tablename = self.get_registry_table_name()
        for uid in uids:
            for level in ["variant", "gene"]:
                q = f"drop table if exists f{level}_{uid}"
                await cursor_write.execute(q)
            q = f"delete from {tablename} where uid=?"
            await cursor_write.execute(q, (uid,))
        await self.conn_write.commit()

    async def drop_ftable(
        self, uid=None, ftype=None, cursor_read=Any, cursor_write=Any
    ):
        if not self.conn_write or not ftype or not uid:
            return
        _ = cursor_read
        table_name = self.get_ftable_name(uid=uid, ftype=ftype)
        q = f"drop table if exists {table_name}"
        await cursor_write.execute(q)
        await self.conn_write.commit()

    async def make_ftables(self):
        if self.should_bypass_filter():
            return None
        if not self.filter:
            return {"uid": None, "status": REPORT_FILTER_NOT_NEEDED}
        ret = await self.exec_db(self.get_existing_report_filter_status)
        if ret:
            self.uid = ret["uid"]
            return ret
        ret = await self.exec_db(self.get_new_report_filter_uid)
        if not ret:
            return None
        [uid, delete_uids] = ret
        if delete_uids:
            for delete_uid in delete_uids:
                await self.exec_db(self.remove_ftables, delete_uid)
        self.uid = uid
        try:
            # register
            await self.exec_db(self.register_new_report_filter, uid=uid)
            # samples to filter
            sample_to_filter = self.get_sample_to_filter()
            if sample_to_filter:
                [req, rej] = sample_to_filter
                await self.exec_db(
                    self.make_sample_to_filter_table, uid=uid, req=req, rej=rej
                )
            # genes to filter
            gene_to_filter = self.get_gene_to_filter()
            if gene_to_filter:
                await self.exec_db(
                    self.make_gene_to_filter_table, uid=uid, genes=gene_to_filter
                )
            await self.make_fvariant(
                uid=uid,
                sample_to_filter=sample_to_filter,
                gene_to_filter=gene_to_filter,
            )
            await self.make_fgene(uid=uid)
            await self.exec_db(
                self.remove_temporary_tables,
                uid=uid,
                gene_to_filter=gene_to_filter,
                sample_to_filter=sample_to_filter,
            )
            await self.exec_db(
                self.set_registry_status, uid=uid, status=REPORT_FILTER_READY
            )
            return {"uid": uid, "status": REPORT_FILTER_READY}
        except Exception as e:
            await self.exec_db(self.remove_ftables, uid)
            raise e

    async def get_variant_data_cols(
        self, cursor_read=Any, cursor_write=Any
    ) -> List[str]:
        _ = cursor_write
        q = f"select * from main.variant limit 1"
        await cursor_read.execute(q)
        return [v[0] for v in cursor_read.description]

    async def get_ftable_num_rows(
        self, level=None, uid=None, ftype=None, cursor_read=Any, cursor_write=Any
    ):
        _ = cursor_write
        if not level or not ftype:
            return
        table_name = self.get_ftable_name(uid=uid, ftype=ftype)
        if table_name:
            q = f"select count(*) from {table_name}"
        else:
            q = f"select count(*) from main.{level}"
        await cursor_read.execute(q)
        ret = await cursor_read.fetchone()
        return ret[0]

    async def get_level_data_iterator(self, level, page=None, pagesize=None, uid=None):
        if not level:
            return None
        if not self.conn_read:
            return None
        ref_col_name = REF_COL_NAMES.get(level)
        if not ref_col_name:
            return None
        cursor_read = await self.conn_read.cursor()
        if not uid:
            filter_uid_status = await self.exec_db(
                self.get_existing_report_filter_status
            )
            if filter_uid_status:
                uid = filter_uid_status.get("uid")
        q = f"select d.* from main.{level} as d"
        if uid:
            ftable = self.get_ftable_name(uid=uid, ftype=level)
            q += f" join {ftable} as f on d.{ref_col_name}=f.{ref_col_name}"
        if page and pagesize:
            offset = (page - 1) * pagesize
            q += f" limit {pagesize} offset {offset}"
        await cursor_read.execute(q)
        rows = await cursor_read.fetchall()
        await cursor_read.close()
        return rows

    async def get_gene_row(self, hugo=None, cursor_read=Any, cursor_write=Any):
        _ = cursor_write
        q = f"select * from main.gene where base__hugo=?"
        await cursor_read.execute(q, (hugo,))
        return await cursor_read.fetchone()

    def get_gene_to_filter(self):
        if isinstance(self.filter, dict) and "genes" in self.filter:
            data = [(hugo,) for hugo in self.filter["genes"]]
        else:
            return None
        if len(data) == 0:
            return None
        return data

    def get_gene_to_filter_table_name(self, uid=None):
        if not uid:
            return None
        return f"{REPORT_FILTER_DB_NAME}.{GENE_TO_FILTER_TABLE_NAME}_{uid}"

    async def make_gene_to_filter_table(
        self, uid=None, genes=None, cursor_read=Any, cursor_write=Any
    ):
        if not self.conn_write:
            return
        _ = cursor_read
        if not uid or not genes:
            return
        table_name = self.get_gene_to_filter_table_name(uid=uid)
        if not table_name:
            return
        q = f"drop table if exists {table_name}"
        await cursor_write.execute(q)
        q = f"create table {table_name} (base__hugo text)"
        await cursor_write.execute(q)
        q = f"insert into {table_name} (base__hugo) values (?)"
        print(f"q={q}")
        await cursor_write.executemany(q, genes)
        await self.conn_write.commit()

    async def make_filtered_hugo_table(
        self, conn=None, cursor_read=Any, cursor_write=Any
    ):
        _ = cursor_read
        bypassfilter = (
            not self.filter
            and not self.filtersql
            and not self.includesample
            and not self.excludesample
        )
        if conn and cursor_write and bypassfilter == False:
            gftable = "gene_filtered"
            q = f"drop table if exists {gftable}"
            await cursor_write.execute(q)
            q = f"create table {gftable} as select distinct v.base__hugo from variant as v inner join variant_filtered as vf on vf.base__uid=v.base__uid where v.base__hugo is not null"
            await cursor_write.execute(q)

    async def table_exists(self, table, cursor_read=Any, cursor_write=Any):
        _ = cursor_write
        sql = (
            'select name from sqlite_master where type="table" and '
            + 'name="'
            + table
            + '"'
        )
        await cursor_read.execute(sql)
        row = await cursor_read.fetchone()
        if row:
            return True
        else:
            return False

    async def get_variant_data_for_cols(self, cols, cursor_read=Any, cursor_write=Any):
        _ = cursor_write
        bypassfilter = self.should_bypass_filter()
        if cols[0] == "base__uid":
            cols[0] = "v.base__uid"
        q = "select {},base__hugo from variant as v".format(",".join(cols))
        if bypassfilter == False:
            q += " inner join variant_filtered as f on v.base__uid=f.base__uid"
        if cols[0] == "v.base__uid":
            cols[0] = "base__uid"
        await cursor_read.execute(q)
        rows = await cursor_read.fetchall()
        return rows

    async def get_result_levels(self, cursor_read=Any, cursor_write=Any) -> List[str]:
        _ = cursor_write
        table_names = []
        q = (
            'select name from sqlite_master where type="table" and '
            + 'name like "%_header"'
        )
        await cursor_read.execute(q)
        for row in await cursor_read.fetchall():
            table_names.append(row[0].replace("_header", ""))
        return table_names

    async def get_stored_output_columns(
        self, module_name, cursor_read=Any, cursor_write=Any
    ):
        from json import loads

        _ = cursor_write
        output_columns = []
        q = f'select col_def from variant_header where col_name like "{module_name}\\_\\_%" escape "\\"'
        await cursor_read.execute(q)
        for row in await cursor_read.fetchall():
            d = loads(row[0])
            d["name"] = d["name"].replace(f"{module_name}__", "")
            output_columns.append(d)
        return output_columns

    async def make_ftables_and_ftable_uid(self, make_filtered_table=True):
        self.ftable_uid = None
        if not self.should_bypass_filter() or make_filtered_table:
            ret = await self.make_ftables()
            if ret:
                ftable_uid = ret["uid"]
                return ftable_uid

    async def getcount(self, level="variant", uid=None):
        if self.should_bypass_filter() or uid:
            norows = await self.exec_db(
                self.get_ftable_num_rows, level=level, uid=uid, ftype=level
            )
        else:
            uid = await self.make_ftables_and_ftable_uid()
            norows = await self.exec_db(
                self.get_ftable_num_rows, level=level, uid=uid, ftype=level
            )
        return norows



loop = None

def get_event_loop():
    from sys import platform as sysplatform
    from asyncio import get_event_loop

    global loop
    if not loop:
        if sysplatform == "win32":  # Required to use asyncio subprocesses
            #from asyncio import ProactorEventLoop  # type: ignore
            import asyncio
            #loop = ProactorEventLoop()
            asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
            loop = get_event_loop()
        else:
            loop = get_event_loop()
    return loop


from typing import List
from typing import Any
from typing import Optional
from ..system.consts import DEFAULT_SERVER_DEFAULT_USERNAME

class BaseReporter:
    def __init__(self, args):
        from ..util.admin_util import get_user_conf

        self.cf = None
        self.filtertable = "filter"
        self.colinfo = {}
        self.colnos = {}
        self.var_added_cols = []
        self.summarizing_modules = []
        self.columngroups = {}
        self.column_subs = {}
        self.column_sub_allow_partial_match = {}
        self.colname_conversion = {}
        self.warning_msgs = []
        self.colnames_to_display = {}
        self.cols_to_display = {}
        self.colnos_to_display = {}
        self.display_select_columns = {}
        self.extracted_cols = {}
        self.conn = None
        self.levels_to_write = None
        self.args = None
        self.dbpath = None
        self.filterpath = None
        self.filtername = None
        self.filterstring = None
        self.filtersql = None
        self.filter = None
        self.output_dir = None
        self.savepath = None
        self.conf = None
        self.module_name = None
        self.module_conf = None
        self.report_types = None
        self.output_basename = None
        self.status_fpath = None
        self.nogenelevelonvariantlevel = None
        self.concise_report = None
        self.extract_columns_multilevel = {}
        self.logger = None
        self.error_logger = None
        self.unique_excs = None
        self.mapper_name = None
        self.level = None
        self.no_log = False
        self.colcount = {}
        self.columns = {}
        self.conns = []
        self.gene_summary_datas = {}
        self.total_norows: Optional[int] = None
        self.module_options: dict = {}
        self.legacy_samples_col = False
        self.priority_colgroupnames = (get_user_conf() or {}).get(
            "report_module_order",
            [
                "base",
                "tagsampler",
                "gencode",
                "hg38",
                "hg19",
                "hg18",
            ],
        )
        self.modules_to_add_to_base = []
        self.user = DEFAULT_SERVER_DEFAULT_USERNAME
        self.parse_cmd_args(args)
        self._setup_logger()

    async def exec_db(self, func, *args, **kwargs):
        from ..exceptions import DatabaseConnectionError

        conn = await self.get_db_conn()
        if not conn:
            raise DatabaseConnectionError(self.module_name)
        cursor = await conn.cursor()
        try:
            ret = await func(*args, conn=conn, cursor=cursor, **kwargs)
        except:
            await cursor.close()
            raise
        await cursor.close()
        return ret

    def parse_cmd_args(self, args):
        import sqlite3
        from os.path import dirname
        from os.path import basename
        from os.path import join
        from os.path import exists
        import json
        from ..module.local import get_module_conf
        from os.path import abspath
        from ..system import consts
        from ..exceptions import WrongInput
        from ..util.run import get_module_options

        if not args:
            return
        if args.get("md"):
            consts.custom_modules_dir = args.get("md")
        self.dbpath = args.get("dbpath")
        if not exists(self.dbpath):
            raise WrongInput(msg=self.dbpath)
        try:
            with sqlite3.connect(self.dbpath) as db:
                db.execute("select * from info")
        except:
            raise WrongInput(msg=f"{self.dbpath} is not an OakVar database")
        self.filterpath = args.get("filterpath")
        self.filtername = args.get("filtername")
        self.filterstring = args.get("filterstring")
        self.filtersql = args.get("filtersql")
        self.filter = args.get("filter")
        self.output_dir = args.get("output_dir")
        self.module_name = args.get("module_name")
        self.report_types = args.get("reports")
        self.savepath = args.get("savepath")
        if self.output_dir:
            self.output_dir = dirname(self.dbpath)
        if not self.output_dir:
            self.output_dir = abspath(".")
        if self.savepath and dirname(self.savepath) == "":
            self.savepath = join(self.output_dir, self.savepath)
        self.module_conf = get_module_conf(self.module_name, module_type="reporter")
        self.conf = args.get("conf")
        self.module_options = get_module_options(args) or {}
        self.confs = self.module_options # TODO: backward compatibility. Delete later.
        self.output_basename = basename(self.dbpath)[:-7]
        self.nogenelevelonvariantlevel = args.get("nogenelevelonvariantlevel", False)
        inputfiles = args.get("inputfiles")
        dbpath = args.get("dbpath")
        if not inputfiles and dbpath:
            db = sqlite3.connect(dbpath)
            c = db.cursor()
            q = 'select colval from info where colkey="_input_paths"'
            c.execute(q)
            r = c.fetchone()
            if r is not None:
                args["inputfiles"] = []
                s = r[0]
                if " " in s:
                    s = s.replace("'", '"')
                s = s.replace("\\", "\\\\\\\\")
                s = json.loads(s)
                for k in s:
                    input_path = s[k]
                    args["inputfiles"].append(input_path)
            c.close()
            db.close()
        self.inputfiles = args.get("inputfiles")
        self.serveradmindb = args.get("serveradmindb")
        self.concise_report = args.get("concise_report")
        if args.get("cols"):
            self.extract_columns_multilevel = {}
            for level in ["variant", "gene", "sample", "mapping"]:
                self.extract_columns_multilevel[level] = args.get("cols")
        else:
            self.extract_columns_multilevel = self.get_standardized_module_option(
                self.module_options.get("extract_columns", {})
            )
        self.add_summary = not args.get("no_summary", False)
        self.args = args

    def should_write_level(self, level):
        if self.levels_to_write is None:
            return True
        elif level in self.levels_to_write:
            return True
        else:
            return False

    async def connect_db(self, dbpath=None):
        _ = dbpath

    async def set_legacy_samples_column_flag(self):
        from sqlite3 import connect

        if not self.dbpath:
            raise
        conn = connect(self.dbpath)
        cursor = conn.cursor()
        q = f"pragma table_info(variant)"
        cursor.execute(q)
        header_cols = [row[1] for row in cursor.fetchall()]
        if "tagsampler__samples" in header_cols:
            self.legacy_samples_col = False
        elif "base__samples" in header_cols:
            self.legacy_samples_col = True
        else:
            raise
        cursor.close()
        conn.close()

    async def prep(self, user=DEFAULT_SERVER_DEFAULT_USERNAME):
        await self.set_dbpath()
        await self.connect_db(dbpath=self.dbpath)
        await self.set_legacy_samples_column_flag()
        await self.load_filter(user=user)

    def _setup_logger(self):
        if self.module_name is None:
            return
        import logging

        if getattr(self, "no_log", False):
            return
        try:
            self.logger = logging.getLogger(self.module_name)
        except Exception as e:
            self._log_exception(e)
        self.error_logger = logging.getLogger("err." + self.module_name)
        self.unique_excs = []

    async def get_db_conn(self):
        import aiosqlite

        if self.dbpath is None:
            return None
        if not self.conn:
            self.conn = await aiosqlite.connect(self.dbpath)
            self.conns.append(self.conn)
        return self.conn

    def _log_exception(self, e, halt=True):
        if halt:
            raise e
        elif self.logger:
            self.logger.exception(e)

    def substitute_val(self, level, row):
        from json import loads
        from json import dumps

        for sub in self.column_subs.get(level, []):
            col_name = f"{sub.module}__{sub.col}"
            value = row[col_name]
            if value is None or value == "" or value == "{}":
                continue
            if (
                level == "variant"
                and sub.module == "base"
                and sub.col == "all_mappings"
            ):
                mappings = loads(value)
                for gene in mappings:
                    for i in range(len(mappings[gene])):
                        sos = mappings[gene][i][2].split(",")
                        sos = [sub.subs.get(so, so) for so in sos]
                        mappings[gene][i][2] = ",".join(sos)
                value = dumps(mappings)
            elif level == "gene" and sub.module == "base" and sub.col == "all_so":
                vals = []
                for i, so_count in enumerate(value.split(",")):
                    so = so_count[:3]
                    so = sub.subs.get(so, so)
                    so_count = so + so_count[3:]
                    vals.append(so_count)
                value = ",".join(vals)
            else:
                value = sub.subs.get(value, value)
            row[col_name] = value
        return row

    def get_extracted_header_columns(self, level):
        cols = []
        for col in self.colinfo[level]["columns"]:
            if col["col_name"] in self.colnames_to_display[level]:
                cols.append(col)
        return cols

    def get_db_col_name(self, mi, col):
        if mi.name in ["gencode", "hg38", "tagsampler"]:
            grp_name = "base"
        else:
            grp_name = mi.name
        return f"{grp_name}__{col['name']}"

    def col_is_categorical(self, col):
        return "category" in col and col["category"] in ["single", "multi"]

    async def do_gene_level_summary(self, add_summary=True):
        _ = add_summary
        self.gene_summary_datas = {}
        if not self.summarizing_modules:
            return self.gene_summary_datas
        for mi, module_instance, summary_cols in self.summarizing_modules:
            gene_summary_data = await module_instance.get_gene_summary_data(self.cf)
            self.gene_summary_datas[mi.name] = [gene_summary_data, summary_cols]
            columns = self.colinfo["gene"]["columns"]
            for col in summary_cols:
                if not self.col_is_categorical(col):
                    continue
                colinfo_col = {}
                colno = None
                for i in range(len(columns)):
                    if columns[i]["col_name"] == self.get_db_col_name(mi, col):
                        colno = i
                        break
                cats = []
                for hugo in gene_summary_data:
                    val = gene_summary_data[hugo][col["name"]]
                    repsub = colinfo_col.get("reportsub", [])
                    if len(repsub) > 0:
                        if val in repsub:
                            val = repsub[val]
                    if val not in cats:
                        cats.append(val)
                if colno is not None:
                    columns[colno]["col_cats"] = cats

    async def store_mapper(self, conn=None, cursor=None):
        from ..exceptions import DatabaseConnectionError

        if conn is None or cursor is None:
            raise DatabaseConnectionError(self.module_name)
        q = 'select colval from info where colkey="_mapper"'
        await cursor.execute(q)
        r = await cursor.fetchone()
        if r is None:
            self.mapper_name = "hg38"
        else:
            self.mapper_name = r[0].split(":")[0]

    def write_log(self, msg):
        if not self.logger:
            return
        self.logger.info(msg)

    def log_run_start(self):
        from time import asctime, localtime
        import oyaml as yaml
        from ..util.run import update_status

        self.write_log("started: %s" % asctime(localtime(self.start_time)))
        if self.cf and self.cf.filter:
            self.write_log(f"filter:\n{yaml.dump(self.filter)}")
        if self.module_conf:
            status = f"started {self.module_conf['title']} ({self.module_name})"
            update_status(status, logger=self.logger, serveradmindb=self.serveradmindb)

    async def get_levels_to_run(self, tab: str) -> List[str]:
        if not self.cf:
            return []
        if tab == "all":
            levels = await self.cf.exec_db(self.cf.get_result_levels)
        else:
            levels = [tab]
        if type(levels) is not list:
            return []
        if not levels:
            return []
        return levels

    async def run(
        self,
        tab="all",
        add_summary=None,
        pagesize=None,
        page=None,
        make_filtered_table=True,
        user=DEFAULT_SERVER_DEFAULT_USERNAME,
        dictrow=False,
    ):
        from ..exceptions import SetupError
        from time import time
        from time import asctime
        from time import localtime
        from ..util.run import update_status

        _ = user
        try:
            # TODO: disabling gene level summary for now. Enable later.
            add_summary = False
            if add_summary is None:
                add_summary = self.add_summary
            self.dictrow = dictrow
            await self.prep()
            if not self.args or not self.cf or not self.logger:
                raise SetupError(self.module_name)
            self.start_time = time()
            ret = None
            tab = tab or self.args.get("level", "all")
            self.log_run_start()
            if self.setup() == False:
                await self.close_db()
                raise SetupError(self.module_name)
            self.ftable_uid = await self.cf.make_ftables_and_ftable_uid(
                make_filtered_table=make_filtered_table
            )
            self.levels = await self.get_levels_to_run(tab)
            for level in self.levels:
                self.level = level
                await self.make_col_infos(add_summary=add_summary)
                await self.write_data(
                    level,
                    pagesize=pagesize,
                    page=page,
                    make_filtered_table=make_filtered_table,
                    add_summary=add_summary,
                )
            await self.close_db()
            if self.module_conf:
                status = "finished {self.module_conf['title']} ({self.module_name})"
                update_status(
                    status, logger=self.logger, serveradmindb=self.serveradmindb
                )
            end_time = time()
            if not (hasattr(self, "no_log") and self.no_log):
                self.logger.info("finished: {0}".format(asctime(localtime(end_time))))
                run_time = end_time - self.start_time
                self.logger.info("runtime: {0:0.3f}".format(run_time))
            ret = self.end()
        except Exception as e:
            await self.close_db()
            import traceback
            traceback.print_exc()
            raise e
        return ret

    async def write_data(
        self,
        level,
        add_summary=True,
        pagesize=None,
        page=None,
        make_filtered_table=True,
    ):
        from ..exceptions import SetupError

        _ = make_filtered_table
        if self.should_write_level(level) == False:
            return
        if not await self.exec_db(self.table_exists, level):
            return
        if not self.cf or not self.args:
            raise SetupError(self.module_name)
        if add_summary and self.level == "gene":
            await self.do_gene_level_summary(add_summary=add_summary)
        self.write_preface(level)
        self.extracted_cols[level] = self.get_extracted_header_columns(level)
        self.write_header(level)
        self.hugo_colno = self.colnos[level].get("base__hugo", None)
        datacols = await self.cf.exec_db(self.cf.get_variant_data_cols)
        self.total_norows = await self.cf.exec_db(self.cf.get_ftable_num_rows, level=level, uid=self.ftable_uid, ftype=level)  # type: ignore
        if datacols is None or self.total_norows is None:
            return
        self.sample_newcolno = None
        if level == "variant" and self.args.get("separatesample"):
            self.write_variant_sample_separately = True
            if self.legacy_samples_col:
                self.sample_newcolno = self.colnos["variant"]["base__samples"]
            else:
                self.sample_newcolno = self.colnos["variant"]["tagsampler__samples"]
        else:
            self.write_variant_sample_separately = False
        datarows_iter = await self.cf.get_level_data_iterator(
            level, page=page, pagesize=pagesize
        )
        if not datarows_iter:
            return
        row_count = 0
        for datarow in datarows_iter:
            datarow = dict(datarow)
            if datarow is None:
                continue
            if level == "gene" and add_summary:
                await self.add_gene_summary_data_to_gene_level(datarow)
            if level == "variant":
                await self.add_gene_level_data_to_variant_level(datarow)
            datarow = self.substitute_val(level, datarow)
            self.stringify_all_mapping(level, datarow)
            self.escape_characters(datarow)
            self.write_row_with_samples_separate_or_not(datarow)
            row_count += 1
            if pagesize and row_count == pagesize:
                break

    def write_row_with_samples_separate_or_not(self, datarow):
        if self.legacy_samples_col:
            col_name = "base__samples"
        else:
            col_name = "tagsampler__samples"
        if self.write_variant_sample_separately:
            samples = datarow[col_name]
            if samples:
                samples = samples.split(";")
                for sample in samples:
                    sample_datarow = datarow
                    sample_datarow[col_name] = sample
                    self.write_table_row(self.get_extracted_row(sample_datarow))
            else:
                self.write_table_row(self.get_extracted_row(datarow))
        else:
            self.write_table_row(self.get_extracted_row(datarow))

    def escape_characters(self, datarow):
        for k, v in datarow.items():
            if isinstance(v, str) and "\n" in v:
                datarow[k] = v.replace("\n", "%0A")

    def stringify_all_mapping(self, level, datarow):
        from json import loads

        if hasattr(self, "keep_json_all_mapping") == True or level != "variant":
            return
        col_name = "base__all_mappings"
        all_map = loads(datarow[col_name])
        newvals = []
        for hugo in all_map:
            for maprow in all_map[hugo]:
                if len(maprow) == 5:
                    # TODO: remove this after a while. Now is 10/22/2022.
                    [protid, protchange, so, transcript, rnachange] = maprow
                    exonno = ""
                else:
                    [protid, protchange, so, transcript, rnachange, exonno] = maprow
                if protid == None:
                    protid = "(na)"
                if protchange == None:
                    protchange = "(na)"
                if rnachange == None:
                    rnachange = "(na)"
                newval = f"{transcript}:{hugo}:{protid}:{so}:{protchange}:{rnachange}:{exonno}"
                newvals.append(newval)
        newvals.sort()
        newcell = "; ".join(newvals)
        datarow[col_name] = newcell

    async def add_gene_summary_data_to_gene_level(self, datarow):
        hugo = datarow["base__hugo"]
        for mi, _, _ in self.summarizing_modules:
            module_name = mi.name
            [gene_summary_data, cols] = self.gene_summary_datas[module_name]
            grp_name = "base" if self.should_be_in_base(module_name) else module_name
            if (
                hugo in gene_summary_data
                and gene_summary_data[hugo] is not None
                and len(gene_summary_data[hugo]) == len(cols)
            ):
                datarow.update(
                    {
                        f"{grp_name}__{col['name']}": gene_summary_data[hugo][
                            col["name"]
                        ]
                        for col in cols
                    }
                )
            else:
                datarow.update({f"{grp_name}__{col['name']}": None for col in cols})

    async def add_gene_level_data_to_variant_level(self, datarow):
        if self.nogenelevelonvariantlevel or self.hugo_colno is None or not self.cf:
            return
        generow = await self.cf.exec_db(self.cf.get_gene_row, datarow["base__hugo"])
        if generow is None:
            datarow.update({col: None for col in self.var_added_cols})
        else:
            datarow.update({col: generow[col] for col in self.var_added_cols})

    async def get_variant_colinfo(self, add_summary=True):
        try:
            await self.prep()
            if self.setup() == False:
                await self.close_db()
                return None
            self.levels = await self.get_levels_to_run("all")
            await self.make_col_infos(add_summary=add_summary)
            return self.colinfo
        except Exception as _:
            import traceback

            traceback.print_exc()
            await self.close_db()
            return None

    def setup(self):
        pass

    def end(self):
        pass

    def write_preface(self, __level__):
        pass

    def write_header(self, __level__):
        pass

    def write_table_row(self, __row__):
        pass

    def get_extracted_row(self, row):
        if self.dictrow:
            filtered_row = {col: row[col] for col in self.cols_to_display[self.level]}
        else:
            filtered_row = [row[col] for col in self.colnames_to_display[self.level]]
        return filtered_row

    def add_to_colnames_to_display(self, level, column):
        """
        include columns according to --cols option
        """
        col_name = column["col_name"]
        if (
            level in self.extract_columns_multilevel
            and len(self.extract_columns_multilevel[level]) > 0
        ):
            if col_name in self.extract_columns_multilevel[level]:
                incl = True
            else:
                incl = False
        elif self.concise_report:
            if "col_hidden" in column and column["col_hidden"] == True:
                incl = False
            else:
                incl = True
        else:
            incl = True
        if incl and col_name not in self.colnames_to_display[level]:
            self.colnames_to_display[level].append(col_name)

    async def make_sorted_column_groups(self, level, conn=Any):
        cursor = await conn.cursor()
        self.columngroups[level] = []
        sql = f"select name, displayname from {level}_annotator order by name"
        await cursor.execute(sql)
        rows = await cursor.fetchall()
        for row in rows:
            (name, displayname) = row
            if name == "base":
                self.columngroups[level].append(
                    {"name": name, "displayname": displayname, "count": 0}
                )
                break
        for row in rows:
            (name, displayname) = row
            if name in self.modules_to_add_to_base:
                self.columngroups[level].append(
                    {"name": name, "displayname": displayname, "count": 0}
                )
        for row in rows:
            (name, displayname) = row
            if name != "base" and name not in self.modules_to_add_to_base:
                self.columngroups[level].append(
                    {"name": name, "displayname": displayname, "count": 0}
                )

    async def make_coldefs(self, level, conn=Any, where=None):
        from ..util.inout import ColumnDefinition

        if not conn:
            return
        cursor = await conn.cursor()
        header_table = f"{level}_header"
        coldefs = []
        sql = f"select col_name, col_def from {header_table}"
        if where:
            sql += f" where {where}"
        await cursor.execute(sql)
        rows = await cursor.fetchall()
        for row in rows:
            col_name, coljson = row
            group_name = col_name.split("__")[0]
            if group_name == "base" or group_name in self.modules_to_add_to_base:
                coldef = ColumnDefinition({})
                coldef.from_json(coljson)
                coldef.level = level
                coldef = await self.gather_col_categories(level, coldef, conn)
                coldefs.append(coldef)
        for row in rows:
            col_name, coljson = row
            group_name = col_name.split("__")[0]
            if group_name == "base" or group_name in self.modules_to_add_to_base:
                continue
            coldef = ColumnDefinition({})
            coldef.from_json(coljson)
            coldef.level = level
            coldef = await self.gather_col_categories(level, coldef, conn)
            coldefs.append(coldef)
        return coldefs

    async def gather_col_categories(self, level, coldef, conn):
        cursor = await conn.cursor()
        if not coldef.category in ["single", "multi"] or len(coldef.categories) > 0:
            return coldef
        sql = f"select distinct {coldef.name} from {level}"
        await cursor.execute(sql)
        rs = await cursor.fetchall()
        for r in rs:
            coldef.categories.append(r[0])
        return coldef

    async def make_columns_colnos_colnamestodisplay_columngroup(self, level, coldefs):
        self.columns[level] = []
        self.colnos[level] = {}
        self.colcount[level] = 0
        for coldef in coldefs:
            self.colnos[level][coldef.name] = self.colcount[level]
            self.colcount[level] += 1
            [colgrpname, _] = self.get_group_field_names(coldef.name)
            column = coldef.get_colinfo()
            self.columns[level].append(column)
            self.add_to_colnames_to_display(level, column)
            for columngroup in self.columngroups[level]:
                if columngroup["name"] == colgrpname:
                    columngroup["count"] += 1

    async def get_gene_level_modules_to_add_to_variant_level(self, conn):
        cursor = await conn.cursor()
        q = "select name from gene_annotator"
        await cursor.execute(q)
        gene_annotators = [v[0] for v in await cursor.fetchall()]
        modules_to_add = [m for m in gene_annotators if m != "base"]
        return modules_to_add

    async def add_gene_level_displayname_to_variant_level_columngroups(
        self, module_name, coldefs, conn
    ):
        cursor = await conn.cursor()
        q = "select displayname from gene_annotator where name=?"
        await cursor.execute(q, (module_name,))
        r = await cursor.fetchone()
        displayname = r[0]
        self.columngroups["variant"].append(
            {"name": module_name, "displayname": displayname, "count": len(coldefs)}
        )

    async def add_gene_level_columns_to_variant_level(self, conn):
        if not await self.exec_db(self.table_exists, "gene"):
            return
        modules_to_add = await self.get_gene_level_modules_to_add_to_variant_level(conn)
        for module_name in modules_to_add:
            module_prefix = f"{module_name}__"
            gene_coldefs = await self.make_coldefs(
                "gene", conn=conn, where=f"col_name like '{module_prefix}%'"
            )
            if not gene_coldefs:
                continue
            await self.add_gene_level_displayname_to_variant_level_columngroups(
                module_name, gene_coldefs, conn
            )
            for gene_coldef in gene_coldefs:
                self.colnos["variant"][gene_coldef.name] = self.colcount["variant"]
                self.colcount["variant"] += 1
                gene_column = gene_coldef.get_colinfo()
                self.columns["variant"].append(gene_column)
                self.add_to_colnames_to_display("variant", gene_column)
                self.var_added_cols.append(gene_coldef.name)

    async def add_gene_level_summary_columns(
        self, add_summary=True, conn=Any, cursor=Any
    ):
        from os.path import dirname
        import sys
        from ..exceptions import ModuleLoadingError
        from ..module.local import get_local_module_infos_of_type
        from ..module.local import get_local_module_info
        from ..util.util import load_class
        from ..util.util import quiet_print
        from ..util.inout import ColumnDefinition

        _ = conn
        if not add_summary:
            return
        q = "select name from variant_annotator"
        await cursor.execute(q)
        done_var_annotators = [v[0] for v in await cursor.fetchall()]
        self.summarizing_modules = []
        local_modules = get_local_module_infos_of_type("annotator")
        local_modules.update(get_local_module_infos_of_type("postaggregator"))
        summarizer_module_names = []
        for module_name in done_var_annotators:
            if module_name == self.mapper_name or module_name in [
                "base",
                "hg38",
                "hg19",
                "hg18",
                "extra_vcf_info",
                "extra_variant_info",
                "original_input",
            ]:
                continue
            if module_name not in local_modules:
                if module_name != "original_input":
                    quiet_print(
                        "            [{}] module does not exist in the system. Gene level summary for this module is skipped.".format(
                            module_name
                        ),
                        self.args,
                    )
                continue
            module = local_modules[module_name]
            if "can_summarize_by_gene" in module.conf:
                summarizer_module_names.append(module_name)
        local_modules[self.mapper_name] = get_local_module_info(self.mapper_name)
        summarizer_module_names = [self.mapper_name] + summarizer_module_names
        for module_name in summarizer_module_names:
            if not module_name:
                continue
            mi = local_modules[module_name]
            if not mi:
                continue
            sys.path = sys.path + [dirname(mi.script_path)]
            annot_cls = None
            if mi.name in done_var_annotators or mi.name == self.mapper_name:
                annot_cls = load_class(mi.script_path)
            if not annot_cls:
                raise ModuleLoadingError(mi.name)
            cmd = {
                "script_path": mi.script_path,
                "input_file": "__dummy__",
                "output_dir": self.output_dir,
                "serveradmindb": self.serveradmindb,
            }
            annot = annot_cls(cmd)
            cols = mi.conf["gene_summary_output_columns"]
            columngroup = {
                "name": mi.name,
                "displayname": mi.title,
                "count": len(cols),
            }
            level = "gene"
            self.columngroups[level].append(columngroup)
            for col in cols:
                coldef = ColumnDefinition(col)
                if self.should_be_in_base(mi.name):
                    coldef.name = f"base__{coldef.name}"
                else:
                    coldef.name = f"{mi.name}__{coldef.name}"
                coldef.genesummary = True
                column = coldef.get_colinfo()
                self.columns[level].append(column)
                self.add_to_colnames_to_display(level, column)
                self.colnos[level][coldef.name] = len(self.colnos[level])
            self.summarizing_modules.append([mi, annot, cols])

    def should_be_in_base(self, name):
        if "__" in name:
            name = self.get_group_name(name)
        return name in self.modules_to_add_to_base

    def get_group_field_names(self, col_name):
        return col_name.split("__")

    def get_group_name(self, col_name):
        return self.get_group_field_names(col_name)[0]

    async def make_report_sub(self, level, conn):
        from json import loads
        from types import SimpleNamespace

        if not level in ["variant", "gene"]:
            return
        reportsubtable = f"{level}_reportsub"
        if not await self.exec_db(self.table_exists, reportsubtable):
            return
        cursor = await conn.cursor()
        q = f"select * from {reportsubtable}"
        await cursor.execute(q)
        reportsub = {r[0]: loads(r[1]) for r in await cursor.fetchall()}
        self.column_subs[level] = []
        for i, column in enumerate(self.columns[level]):
            module_name, field_name = self.get_group_field_names(column["col_name"])
            if module_name == self.mapper_name:
                module_name = "base"
            if module_name in reportsub and field_name in reportsub[module_name]:
                self.column_subs[level].append(
                    SimpleNamespace(
                        module=module_name,
                        col=field_name,
                        index=i,
                        subs=reportsub[module_name][field_name],
                    )
                )
                self.columns[level][i]["reportsub"] = reportsub[module_name][field_name]

    def set_display_select_columns(self, level):
        if self.extract_columns_multilevel.get(level, {}) or self.concise_report:
            self.display_select_columns[level] = True
        else:
            self.display_select_columns[level] = False

    def set_cols_to_display(self, level):
        self.cols_to_display[level] = []
        self.colnos_to_display[level] = []
        colno = 0
        for col in self.columns[level]:
            col_name = col["col_name"]
            if col_name in self.colnames_to_display[level]:
                self.cols_to_display[level].append(col_name)
                self.colnos_to_display[level].append(colno)
            colno += 1

    async def make_col_infos(self, add_summary=True):
        prev_level = self.level
        for level in self.levels:
            self.level = level
            await self.exec_db(self.make_col_info, level, add_summary=add_summary)
        self.level = prev_level

    def add_column_number_stat_to_col_groups(self, level: str):
        last_columngroup_pos = 0
        for columngroup in self.columngroups.get(level, []):
            columngroup["start_column_number"] = last_columngroup_pos
            new_last_columngroup_pos = last_columngroup_pos + columngroup["count"]
            columngroup["end_colunm_number"] = new_last_columngroup_pos
            last_columngroup_pos = new_last_columngroup_pos

    async def make_col_info(self, level: str, add_summary=True, conn=Any, cursor=Any):
        _ = cursor
        if not level or not await self.exec_db(self.table_exists, level):
            return
        await self.exec_db(self.store_mapper)
        self.colnames_to_display[level] = []
        self.modules_to_add_to_base = [self.mapper_name, "tagsampler"]
        await self.make_sorted_column_groups(level, conn=conn)
        coldefs = await self.make_coldefs(level, conn=conn)
        if not coldefs:
            return
        await self.make_columns_colnos_colnamestodisplay_columngroup(level, coldefs)
        if not self.nogenelevelonvariantlevel and self.level == "variant":
            await self.add_gene_level_columns_to_variant_level(conn)
        if self.level == "gene" and level == "gene" and add_summary:
            await self.exec_db(self.add_gene_level_summary_columns, level)
        self.set_display_select_columns(level)
        self.set_cols_to_display(level)
        self.add_column_number_stat_to_col_groups(level)
        self.colinfo[level] = {
            "colgroups": self.columngroups[level],
            "columns": self.columns[level],
        }
        await self.make_report_sub(level, conn)

    def get_standardized_module_option(self, v):
        tv = type(v)
        if tv == str:
            if ":" in v:
                v0 = {}
                for v1 in v.split("."):
                    if ":" in v1:
                        v1toks = v1.split(":")
                        if len(v1toks) == 2:
                            level = v1toks[0]
                            v2s = v1toks[1].split(",")
                            v0[level] = v2s
                v = v0
            elif "," in v:
                v = [val for val in v.split(",") if val != ""]
        if v == "true":
            v = True
        elif v == "false":
            v = False
        return v

    async def set_dbpath(self, dbpath=None):
        from os.path import exists
        from ..exceptions import NoInput
        from ..exceptions import WrongInput

        if dbpath != None:
            self.dbpath = dbpath
        if not self.dbpath:
            raise NoInput()
        if not exists(self.dbpath):
            raise WrongInput()

    async def close_db(self):
        import sqlite3

        for conn in self.conns:
            if type(conn) == sqlite3.Connection:
                conn.close()
            else:
                await conn.close()
        self.conns = []
        if self.cf is not None:
            await self.cf.close_db()
            self.cf = None

    async def load_filter(self, user=DEFAULT_SERVER_DEFAULT_USERNAME):
        from ..exceptions import SetupError
        from .. import ReportFilter

        if self.args is None:
            raise SetupError()
        self.cf = await ReportFilter.create(dbpath=self.dbpath, user=user, strict=False)
        await self.cf.exec_db(
            self.cf.loadfilter,
            filter=self.filter,
            filterpath=self.filterpath,
            filtername=self.filtername,
            filterstring=self.filterstring,
            filtersql=self.filtersql,
            includesample=self.args.get("includesample"),
            excludesample=self.args.get("excludesample"),
        )

    async def table_exists(self, tablename, conn=None, cursor=None):
        if conn is None:
            pass
        if cursor is None:
            from ..exceptions import SetupError

            raise SetupError()
        sql = (
            "select name from sqlite_master where "
            + 'type="table" and name="'
            + tablename
            + '"'
        )
        await cursor.execute(sql)
        row = await cursor.fetchone()
        if row == None:
            ret = False
        else:
            ret = True
        return ret

CravatReport = BaseReporter



from typing import Optional

class BaseMapper(object):
    def __init__(self, *inargs, **inkwargs):
        from time import time
        from pathlib import Path
        import pkg_resources
        from pathlib import Path
        from ..module.local import get_module_conf

        self.cmd_parser = None
        self.input_path: Optional[Path] = None
        self.input_dir = None
        self.output_dir: Optional[Path] = None
        self.reader = None
        self.output_base_fname = None
        self.crx_path = None
        self.crg_path = None
        self.crx_writer = None
        self.crg_writer = None
        self.input_fname = None
        self.module_options = None
        self.primary_transcript_paths = None
        self.args = None
        self.logger = None
        self.error_logger = None
        self.unique_excs = []
        self.written_primary_transc = None
        self.define_main_cmd_args()
        self.parse_args(inargs, inkwargs)
        if self.args is None:
            return
        self.t = time()
        self.serveradmindb = self.args.get("serveradmindb")
        main_fpath = self.args.get("script_path", __file__)
        self.module_name = Path(main_fpath).stem
        self.module_dir = Path(main_fpath).parent
        self.mapper_dir = self.module_dir
        self.gene_sources = []
        self.gene_info = {}
        self.setup_logger()
        self.conf = get_module_conf(self.module_name, module_type="mapper")
        self.cravat_version = pkg_resources.get_distribution("oakvar").version

    def define_main_cmd_args(self):
        import argparse

        self.cmd_parser = argparse.ArgumentParser()
        self.cmd_parser.add_argument("input_file", help="Input crv file")
        self.cmd_parser.add_argument(
            "-n", dest="run_name", help="Name of job. " + "Default is input file name."
        )
        self.cmd_parser.add_argument(
            "-d",
            dest="output_dir",
            help="Output directory. " + "Default is input file directory.",
        )
        self.cmd_parser.add_argument(
            "--confs", dest="confs", default="{}", help="Configuration string"
        )
        self.cmd_parser.add_argument(
            "--seekpos", dest="seekpos", default=None, help=argparse.SUPPRESS
        )
        self.cmd_parser.add_argument(
            "--chunksize", dest="chunksize", default=None, help=argparse.SUPPRESS
        )
        self.cmd_parser.add_argument(
            "--primary-transcript",
            dest="primary_transcript",
            # nargs="*",
            default=["mane"],
            help='"mane" for MANE transcripts as primary transcripts, or a path to a file of primary transcripts. MANE is default.',
        )

    def parse_args(self, inargs, inkwargs):
        from os import makedirs
        from pathlib import Path
        from ..util.util import get_args
        from ..util.run import get_module_options
        from ..consts import STANDARD_INPUT_FILE_SUFFIX

        args = get_args(self.cmd_parser, inargs, inkwargs)
        self.input_path = args.get("input_file") or None
        if self.input_path:
            p = Path(self.input_path).absolute()
            self.input_dir = p.parent
            self.input_fname = p.name
            if args.get("output_dir"):
                self.output_dir = Path(args.get("output_dir"))
            else:
                self.output_dir = self.input_dir
            if not self.output_dir.exists():
                makedirs(self.output_dir)
        self.output_base_fname = args.get("run_name")
        if not self.output_base_fname and self.input_fname:
            self.output_base_fname = self.input_fname
            p = Path(self.output_base_fname)
            if p.suffix == STANDARD_INPUT_FILE_SUFFIX:
                self.output_base_fname = p.stem
        self.module_options = get_module_options(args)
        self.primary_transcript_paths = [v for v in args["primary_transcript"] if v]
        self.postfix = args.get("postfix") or ""
        self.args = args

    def setup(self):
        raise NotImplementedError("Mapper must have a setup() method.")

    def extra_setup(self):
        pass

    def end(self):
        pass

    def setup_logger(self):
        from logging import getLogger

        self.logger = getLogger("oakvar.mapper")
        if self.input_path:
            self.logger.info("input file: %s" % self.input_path)
        self.error_logger = getLogger("err." + self.module_name)

    def make_input_reader(self):
        from ..util.inout import FileReader

        if (
            self.args is not None
            and self.args["seekpos"] is not None
            and self.args["chunksize"] is not None
        ):
            self.reader = FileReader(
                self.input_path,
                seekpos=int(self.args["seekpos"]),
                chunksize=int(self.args["chunksize"]),
            )
        else:
            self.reader = FileReader(self.input_path)
        return self.reader

    def make_crx_writer(self):
        from ..util.inout import FileWriter
        from ..util.util import get_crx_def
        from ..consts import crx_idx
        from ..consts import VARIANT_LEVEL_MAPPED_FILE_SUFFIX

        if not self.output_dir or self.conf is None or not self.args:
            raise
        crx_def = get_crx_def()
        crx_fname = f"{self.output_base_fname}{VARIANT_LEVEL_MAPPED_FILE_SUFFIX}"
        self.crx_path = self.output_dir / (crx_fname + self.postfix)
        self.crx_writer = FileWriter(self.crx_path)
        self.crx_writer.add_columns(crx_def)
        self.crx_writer.write_definition(self.conf)
        for index_columns in crx_idx:
            self.crx_writer.add_index(index_columns)
        self.crx_writer.write_meta_line("title", self.conf["title"])
        self.crx_writer.write_meta_line("version", self.conf["version"])
        self.crx_writer.write_meta_line("modulename", self.module_name)
        if not self.primary_transcript_paths:
            self.crx_writer.write_meta_line("primary_transcript_paths", "")
        else:
            for path in self.primary_transcript_paths:
                self.crx_writer.write_meta_line(
                    "primary_transcript_path", path
                )

    def make_crg_writer(self):
        from ..util.util import get_crg_def
        from ..util.inout import FileWriter
        from ..consts import GENE_LEVEL_MAPPED_FILE_SUFFIX
        from ..consts import crg_idx

        if not self.output_dir or not self.args:
            raise
        crg_def = get_crg_def()
        crg_fname = f"{self.output_base_fname}{GENE_LEVEL_MAPPED_FILE_SUFFIX}"
        self.crg_path = self.output_dir / (crg_fname + self.postfix)
        self.crg_writer = FileWriter(self.crg_path)
        self.crg_writer.add_columns(crg_def)
        self.crg_writer.write_definition(self.conf)
        for index_columns in crg_idx:
            self.crg_writer.add_index(index_columns)

    def setup_input_output(self):
        self.make_input_reader()
        self.make_crx_writer()
        self.make_crg_writer()

    def map(self, crv_data: dict):
        _ = crv_data

    def process_file(self):
        from time import time
        from ..util.run import update_status

        if not self.reader or not self.crx_writer or not self.args:
            raise
        count = 0
        last_status_update_time = time()
        crx_data = None
        for ln, line, crv_data in self.reader.loop_data():
            try:
                count += 1
                cur_time = time()
                if count % 10000 == 0 or cur_time - last_status_update_time > 3:
                    status = f"Running gene mapper: line {count}"
                    update_status(
                        status, logger=self.logger, serveradmindb=self.serveradmindb
                    )
                    last_status_update_time = cur_time
                if crv_data["alt_base"] == "*":
                    crx_data = crv_data
                    crx_data["all_mappings"] = "{}"
                else:
                    crx_data = self.map(crv_data)
            except Exception as e:
                self.log_runtime_error(ln, line, e, fn=self.reader.path)
            if crx_data:
                self.crx_writer.write_data(crx_data)
                self.add_crx_to_gene_info(crx_data)

    def add_crx_to_gene_info(self, crx_data):
        from ..util.inout import AllMappingsParser

        tmap_json = crx_data["all_mappings"]
        # Return if no tmap
        if tmap_json == "":
            return
        tmap_parser = AllMappingsParser(tmap_json)
        for hugo in tmap_parser.get_genes():
            self.gene_info[hugo] = True

    def write_crg(self):
        from ..util.util import get_crg_def

        if self.crg_writer is None:
            return
        sorted_hugos = list(self.gene_info.keys())
        sorted_hugos.sort()
        for hugo in sorted_hugos:
            crg_data = {x["name"]: "" for x in get_crg_def()}
            crg_data["hugo"] = hugo
            self.crg_writer.write_data(crg_data)

    def log_runtime_error(self, ln, line, e, fn=None):
        import traceback

        _ = line
        err_str = traceback.format_exc().rstrip()
        if (
            self.logger is not None
            and self.unique_excs is not None
            and err_str not in self.unique_excs
        ):
            self.unique_excs.append(err_str)
            self.logger.error(err_str)
        if self.error_logger is not None:
            self.error_logger.error(f"{fn}:{ln}\t{str(e)}")

    async def get_gene_summary_data(self, cf):
        from ..util.util import get_crx_def
        from json import loads
        from ..gui.consts import result_viewer_num_var_limit_for_gene_summary_key
        from ..gui.consts import DEFAULT_RESULT_VIEWER_NUM_VAR_LIMIT_FOR_GENE_SUMMARY
        from ..system import get_system_conf

        cols = [
            "base__" + coldef["name"]
            for coldef in get_crx_def()
            if not coldef["name"] in ["cchange", "exonno"]
        ]
        cols.extend(["tagsampler__numsample"])
        data = {}
        rows = await cf.exec_db(cf.get_variant_data_for_cols, cols)
        sys_conf = get_system_conf()
        result_viewer_num_var_limit_for_gene_summary = sys_conf.get(
            result_viewer_num_var_limit_for_gene_summary_key,
            DEFAULT_RESULT_VIEWER_NUM_VAR_LIMIT_FOR_GENE_SUMMARY,
        )
        if len(rows) > result_viewer_num_var_limit_for_gene_summary:
            return {}
        rows_by_hugo = {}
        for row in rows:
            all_mappings = loads(row["base__all_mappings"])
            for hugo in all_mappings.keys():
                if hugo not in rows_by_hugo:
                    rows_by_hugo[hugo] = []
                rows_by_hugo[hugo].append(row)
        hugos = await cf.exec_db(cf.get_filtered_hugo_list)
        for hugo in hugos:
            rows = rows_by_hugo[hugo]
            input_data = {}
            for i in range(len(cols)):
                input_data[cols[i]] = [row[i] for row in rows]
            if hasattr(self, "summarize_by_gene"):
                out = self.summarize_by_gene(hugo, input_data)  # type: ignore
                data[hugo] = out
        return data

    def live_report_substitute(self, d):
        import re

        if self.conf is None or "report_substitution" not in self.conf:
            return
        rs_dic = self.conf["report_substitution"]
        rs_dic_keys = list(rs_dic.keys())
        for colname in d.keys():
            if colname in rs_dic_keys:
                value = d[colname]
                if colname in ["all_mappings", "all_so"]:
                    for target in list(rs_dic[colname].keys()):
                        value = re.sub(
                            "\\b" + target + "\\b", rs_dic[colname][target], value
                        )
                else:
                    if value in rs_dic[colname]:
                        value = rs_dic[colname][value]
                d[colname] = value
        return d

    def run(self, __pos_no__):
        from time import time, asctime, localtime
        from ..util.run import update_status

        self.setup()
        self.setup_input_output()
        self.extra_setup()
        if (
            self.args is None
            or self.logger is None
            or self.conf is None
            or self.reader is None
            or self.crx_writer is None
        ):
            raise
        start_time = time()
        tstamp = asctime(localtime(start_time))
        self.logger.info(f"started: {tstamp} | {self.args['seekpos']}")
        status = f"started {self.conf['title']} ({self.module_name})"
        update_status(status, logger=self.logger, serveradmindb=self.serveradmindb)
        self.process_file()
        self.write_crg()
        stop_time = time()
        tstamp = asctime(localtime(stop_time))
        self.logger.info(f"finished: {tstamp} | {self.args['seekpos']}")
        runtime = stop_time - start_time
        self.logger.info("runtime: %6.3f" % runtime)
        self.end()




from typing import Tuple
from typing import Optional


def get_valid_email_pw(args=None, pwconfirm=False) -> Tuple:
    from ....util.util import email_is_valid
    from ....util.util import pw_is_valid
    from ....exceptions import WrongInput
    from ....system import get_user_conf
    from ....store.consts import ov_store_email_key
    from ....store.consts import ov_store_pw_key
    from getpass import getpass

    if not args:
        raise WrongInput()
    email = args.get("email")
    pw = args.get("pw")
    newpw = args.get("newpw")
    if not email or not pw:
        user_conf = get_user_conf()
        if not email:
            email = user_conf.get(ov_store_email_key)
        if not pw:
            pw = user_conf.get(ov_store_pw_key)
    if not email:
        email = input("Email: ")
    if not pw:
        if pwconfirm:
            pwagain = None
            while not pw or pw != pwagain:
                while not pw:
                    pw = getpass("Password (alphabets, numbers, and !?&@-+): ")
                pwagain = getpass("Confirm password: ")
                if pw != pwagain:
                    print("Password mismatch")
                    pw = None
                    pwagain = None
        else:
            pw = getpass()
    if not email or not pw:
        raise WrongInput("no email or password")
    if not email_is_valid(email):
        raise WrongInput("invalid email")
    if not pw_is_valid(pw):
        raise WrongInput("invalid password")
    if newpw:
        if not pw_is_valid(newpw):
            raise WrongInput("invalid new password")
    if args.get("newpw"):
        return email, pw, newpw
    else:
        return email, pw


def create(email=None, pw=None, args={}, quiet=None) -> dict:
    from requests import post
    from ....system import get_system_conf
    from ....util.util import quiet_print
    from ....util.util import get_email_pw_from_input
    from ....store.consts import store_url_key

    if not email or not pw:
        email = args.get("email")
        pw = args.get("pw")
        if not email or not pw:
            email, pw = get_email_pw_from_input(email=email, pw=pw, pwconfirm=True)
    if not email:
        return {"msg": "No email", "success": False}
    sys_conf = get_system_conf()
    store_url = sys_conf[store_url_key]
    create_account_url = store_url + "/account/create"
    params = {
        "email": email,
        "pw": pw,
    }
    try:
        r = post(create_account_url, data=params)
        status_code = r.status_code
        if status_code == 403:
            msg = "account-exists"
            quiet_print(msg, args=args, quiet=quiet)
            return {"msg": msg, "success": False}
        elif status_code == 202:
            msg = f"Check your inbox for a verification email."
            quiet_print(msg, args=args, quiet=quiet)
            return {"msg": msg, "success": True}
        elif status_code == 201:
            msg = (
                f"Account has been created. Check your inbox for a verification email."
            )
            quiet_print(msg, args=args, quiet=quiet)
            return {"msg": msg, "success": True}
        else:
            msg = f"fail. {r.text}"
            quiet_print(msg, args=args, quiet=quiet)
            return {"msg": msg, "success": False}
    except Exception as e:
        msg = f"fail. {e}"
        quiet_print(msg, args=args, quiet=quiet)
        return {"msg": msg, "success": False}


def delete(args={}) -> bool:
    from requests import post
    from ...ov import get_store_url
    from ....util.util import quiet_print

    token_set = get_token_set()
    if not token_set:
        quiet_print(f"log in first", args=args)
        return False
    store_url = get_store_url()
    url = store_url + "/account/delete"
    params = {"idToken": token_set["idToken"]}
    r = post(url, data=params)
    status_code = r.status_code
    if status_code == 200:
        quiet_print(f"success", args=args)
        return True
    else:
        quiet_print(f"fail. {r.text}", args=args)
        return False


def check(args={}) -> bool:
    from ....util.util import quiet_print

    id_token = get_id_token()
    if not id_token:
        quiet_print(f"not logged in", args=args)
        return False
    valid, expired = id_token_is_valid()
    if valid:
        if expired:
            refresh_token_set()
        token_set = get_token_set() or {}
        email = token_set["email"]
        quiet_print(f"logged in as {email}", args=args)
        return True
    else:
        quiet_print(f"not logged in", args=args)
        return False


def reset(args={}) -> bool:
    from ....util.util import quiet_print
    from ...ov import get_store_url
    from ....util.util import email_is_valid
    from requests import post

    email = args.get("email")
    if not email:
        return False
    if not email_is_valid(email):
        quiet_print(f"invalid email", args=args)
        return False
    url = get_store_url() + "/account/reset"
    params = {"email": email}
    res = post(url, data=params)
    if res.status_code == 200:
        quiet_print(
            "Success. Check your email for instruction to reset your password.",
            args=args,
        )
        return True
    else:
        quiet_print(f"fail. {res.text}", args=args)
        return False


def get_email_from_token_set() -> Optional[str]:
    token_set = get_token_set() or {}
    email = token_set.get("email")
    return email


def login(email=None, pw=None, args={}, quiet=None) -> bool:
    from requests import post
    from ....util.util import quiet_print
    from ....util.util import get_email_from_args
    from ...ov import get_store_url

    id_token = get_id_token()
    if id_token:
        if check(args={"quiet": True}):
            token_set_email = get_email_from_token_set()
            args_email = get_email_from_args(args=args)
            if email:
                if token_set_email == email:
                    quiet_print(f"logged in as {email}", args=args, quiet=quiet)
                    return True
            elif args_email:
                if token_set_email == args_email:
                    quiet_print(f"logged in as {args_email}", args=args, quiet=quiet)
                    return True
    if not email or not pw:
        email, pw = get_valid_email_pw(args=args)
    change_account_url = get_store_url() + "/account/login"
    params = {"email": email, "pw": pw}
    try:
        r = post(change_account_url, data=params)
        status_code = r.status_code
        if status_code == 200:
            save_token_set(r.json())
            quiet_print(f"logged in as {email}", args=args, quiet=quiet)
            return True
        else:
            quiet_print(f"fail. {r.text}", args=args, quiet=quiet)
            return False
    except:
        quiet_print(f"server error", args=args, quiet=quiet)
        return False


def get_token_set_path():
    from os.path import join
    from ....system import get_user_conf_dir
    from ....store.consts import ov_store_id_token_fname

    user_conf_dir = get_user_conf_dir()
    token_path = join(user_conf_dir, ov_store_id_token_fname)
    return token_path


def get_token_set() -> Optional[dict]:
    from os.path import exists
    from json import load

    token_set_path = get_token_set_path()
    if not token_set_path or not exists(token_set_path):
        return None
    with open(token_set_path) as f:
        return load(f)


def get_id_token() -> Optional[str]:
    token_set = get_token_set()
    if not token_set:
        return None
    else:
        return token_set["idToken"]


def get_refresh_token() -> Optional[str]:
    token_set = get_token_set()
    if not token_set:
        return None
    else:
        return token_set["refreshToken"]


def save_token_set(token_set: dict):
    from json import dump

    for k in list(token_set.keys()):
        if "_" in k:
            words = k.split("_")
            newk = words[0] + words[1].capitalize()
            token_set[newk] = token_set[k]
            del token_set[k]
    token_path = get_token_set_path()
    with open(token_path, "w") as wf:
        dump(token_set, wf)


def delete_id_token(args={}):
    from os import remove
    from os.path import exists
    from ....util.util import quiet_print

    token_path = get_token_set_path()
    if exists(token_path):
        remove(token_path)
        return True
    else:
        quiet_print(f"not logged in", args=args)
        return False


def id_token_is_valid() -> Tuple[bool, bool]:  # valid, expired
    from ...ov import get_store_url
    from requests import post

    id_token = get_id_token()
    if not id_token:
        return False, True
    params = {"idToken": id_token}
    url = get_store_url() + "/account/id_token_verified"
    res = post(url, data=params)
    st = res.status_code
    if st == 460:  # valid but expired
        return True, True
    elif st == 200:
        return True, False
    else:
        return False, True


def refresh_token_set() -> bool:
    from ...ov import get_store_url
    from requests import post

    refresh_token = get_refresh_token()
    url = get_store_url() + "/account/refresh"
    params = {"refreshToken": refresh_token}
    res = post(url, data=params)
    if res.status_code == 200:
        token_set = get_token_set()
        if token_set:
            j = res.json()
            id_token = j["id_token"]
            refresh_token = j["refresh_token"]
            token_set["idToken"] = id_token
            token_set["refreshToken"] = refresh_token
            save_token_set(token_set)
            return True
        else:
            return False
    else:
        return False


def change(args={}) -> bool:
    from ....util.util import quiet_print
    from requests import post
    from ...ov import get_store_url
    from getpass import getpass
    from ....util.util import pw_is_valid

    id_token = get_id_token()
    if not id_token:
        quiet_print(f"not logged in", args=args)
        return False
    valid, expired = id_token_is_valid()
    if valid and not expired:
        id_token = None
        refresh_token = get_refresh_token()
        if refresh_token:
            if refresh_token_set():
                token_set = get_token_set()
                if token_set:
                    id_token = token_set["idToken"]
        if not id_token:
            quiet_print(f"not logged in", args=args)
            return False
    newpw = args.get("newpw")
    if not newpw:
        newpw = ""
        while not pw_is_valid(newpw):
            newpw = getpass("New password: ")
    refresh_token = get_refresh_token()
    url = get_store_url() + "/account/change"
    params = {"idToken": id_token, "refreshToken": refresh_token, "newpw": newpw}
    res = post(url, data=params)
    status_code = res.status_code
    if status_code != 200:
        quiet_print(f"{res.text}", args=args)
        return False
    else:
        token_set = get_token_set()
        if not token_set:
            quiet_print(f"password changed but re-login failed")
            return False
        email = token_set["email"]
        args["email"] = email
        args["pw"] = newpw
        if login(args=args):
            return True
        else:
            quiet_print(f"password changed but re-login failed", args=args)
            return True


def logout(args={}) -> bool:
    from ....util.util import quiet_print

    ret = delete_id_token(args=args)
    if ret:
        quiet_print(f"success", args=args)
    return ret


def get_current_id_token(args={}) -> Optional[str]:
    token_set = get_current_token_set(args=args)
    if token_set:
        return token_set["idToken"]
    else:
        return None


def get_current_token_set(args={}) -> Optional[dict]:
    token_set = None
    token_set = get_token_set()
    newargs = args.copy()
    newargs["quiet"] = True
    if token_set:
        valid, expired = id_token_is_valid()
        if not valid:
            token_set = None
        elif expired:
            refresh_token_set()
            token_set = get_token_set()
    if not token_set:
        login(args=newargs)
        token_set = get_token_set()
    return token_set


def token_set_exists() -> bool:
    token_set = get_token_set()
    if token_set:
        return True
    else:
        return False


def delete_token_set():
    from os.path import exists
    from os import remove

    token_set_path = get_token_set_path()
    if exists(token_set_path):
        remove(token_set_path)


def get_email_pw_from_settings(
    email=None, pw=None, conf=None, args={}
) -> Optional[Tuple[str, str]]:
    from ...consts import ov_store_email_key
    from ...consts import ov_store_pw_key
    from ....system import get_user_conf

    # if not given directly, check direct arguments.
    if not email or not pw:
        email = args.get("email")
        pw = args.get("pw")
    # if not, use conf.
    if (not email or not pw) and conf:
        email = conf.get(ov_store_email_key)
        pw = conf.get(ov_store_pw_key)
    # if not, oakvar.yml
    user_conf = get_user_conf()
    if ov_store_email_key in user_conf and ov_store_pw_key in user_conf:
        email = user_conf[ov_store_email_key]
        pw = user_conf[ov_store_pw_key]
    if email and pw:
        return email, pw
    else:
        return None


def emailpw_are_valid(emailpw: Tuple[str, str]) -> bool:
    from ....util.util import email_is_valid
    from ....util.util import pw_is_valid

    email = emailpw[0]
    pw = emailpw[1]
    return email_is_valid(email) and pw_is_valid(pw)


def email_is_verified(email: str, args={}, quiet=None) -> bool:
    from ....util.util import quiet_print
    from ...ov import get_store_url
    from requests import post

    url = get_store_url() + "/account/email_verified"
    params = {"email": email}
    res = post(url, data=params)
    if res.status_code == 200:
        return True
    elif res.status_code == 404:
        quiet_print(f"user not found", args=args, quiet=quiet)
        return False
    else:
        quiet_print(f"{email} has not been verified. {res.text}", args=args, quiet=quiet)
        return False


def announce_on_email_verification_if_needed(email: str, args={}, quiet=None):
    from ....system import show_email_verify_action_banner

    if not email_is_verified(email, args=args, quiet=quiet):
        show_email_verify_action_banner()


def login_with_token_set(args={}) -> bool:
    from ....util.util import quiet_print

    token_set = get_token_set()
    if token_set:
        email = token_set["email"]
        correct, expired = id_token_is_valid()
        email_verified = email_is_verified(email, args=args)
        if not email_verified:
            quiet_print(
                f"Email not verified. A verification email should have been sent to your inbox."
            )
            return True
        else:
            if correct:
                if expired:
                    if refresh_token_set():
                        return True
                    else:
                        delete_token_set()
                else:
                    return True
            else:
                delete_token_set()
    return False


def login_with_email_pw(email=None, pw=None, args={}, conf={}) -> bool:
    emailpw = get_email_pw_from_settings(email=email, pw=pw, args=args, conf=conf)
    if emailpw:
        if emailpw_are_valid(emailpw):
            email = emailpw[0]
            pw = emailpw[1]
            while True:
                ret = create(email=email, pw=pw, quiet=False)
                if ret.get("success"):
                    break
            announce_on_email_verification_if_needed(email, args=args)
            login(email=email, pw=pw, args=args)
            return True
    return False


def total_login(email=None, pw=None, args={}, conf=None) -> bool:
    from ....util.util import get_email_pw_from_input
    from ....system import show_no_user_account_prelude

    if login_with_token_set(args=args):
        return True
    if login_with_email_pw(email=email, pw=pw, args=args, conf=conf):
        return True
    # if not already logged in nor email and pw in settings did not work, get manual input.
    show_no_user_account_prelude()
    while True:
        email, pw = get_email_pw_from_input(pwconfirm=True)
        ret = create(email=email, pw=pw, quiet=False)
        if ret.get("success"):
            break
    announce_on_email_verification_if_needed(email, args=args)
    ret = login(email=email, pw=pw, args=args)
    return ret



from . import cli_entry
from . import cli_func


@cli_entry
def cli_version(args):
    return version(args)


@cli_func
def version(args, __name__="version"):
    from ..util.admin_util import oakvar_version
    from ..util.util import quiet_print

    ret = oakvar_version()
    if args["to"] == "stdout":
        quiet_print(ret, args=args)
    else:
        return ret


def get_parser_fn_version():
    from argparse import ArgumentParser

    # shows version
    parser_cli_version = ArgumentParser()
    parser_cli_version.add_argument(
        "--to", default="return", help='"stdout" to print. "return" to return'
    )
    parser_cli_version.add_argument(
        "--quiet", action="store_true", default=None, help="run quietly"
    )
    parser_cli_version.set_defaults(func=cli_version)
    return parser_cli_version
